/*******************************************************************************
 * Copyright (C) 2010, Jens Baumgart <jens.baumgart@sap.com>
 * Copyright (C) 2008, Robin Rosenberg <robin.rosenberg@dewire.com>
 * Copyright (C) 2008, Shawn O. Pearce <spearce@spearce.org>
 * Copyright (C) 2008, Google Inc.
 * Copyright (C) 2016, Thomas Wolf <thomas.wolf@paranor.ch>
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *******************************************************************************/
package fse.eclipse.branchrecorder.changerecorder;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Path;
import org.eclipse.jgit.lib.Repository;
import org.eclipse.jgit.storage.file.FileRepositoryBuilder;

import fse.eclipse.branchrecorder.util.RepositoryUtil;

/**
 * Central cache for Repository instances.<br>
 * <br>
 * Original source code is {@link org.eclipse.egit.core.RepositoryCache}
 */
public class RepositoryCache {
    private static final List<Repository> repositoryCache = new ArrayList<>();

    public static boolean contains(Repository repository) {
        String path = repository.getDirectory().getAbsolutePath();
        for (Repository r : repositoryCache) {
            String path2 = r.getDirectory().getAbsolutePath();
            if (path.equals(path2)) {
                return true;
            }
        }
        return false;
    }

    /**
     * @return all Repository instances contained in the cache
     */
    public static synchronized Repository[] getAllRepositories() {
        return repositoryCache.toArray(new Repository[repositoryCache.size()]);
    }

    /**
     * Lookup the closest git repository with a working tree containing the given resource. If there are repositories
     * nested above in the file system hierarchy we select the closest one above the given resource.
     *
     * @param resource the resource to find the repository for
     * @return the git repository which has the given resource in its working tree, or null if none found
     * @since 3.2
     */
    public static Repository getRepository(final IResource resource) {
        IPath location = resource.getLocation();
        if (location == null)
            return null;
        return getRepository(location);
    }

    /**
     * Lookup the closest git repository with a working tree containing the given file location. If there are
     * repositories nested above in the file system hierarchy we select the closest one above the given location.
     *
     * @param location the file location to find the repository for
     * @return the git repository which has the given location in its working tree, or null if none found
     * @since 3.2
     */
    public static Repository getRepository(final IPath location) {
        if (location == null) {
            return null;
        }

        Repository[] repositories = getAllRepositories();
        Repository repository = null;
        int largestSegmentCount = 0;
        for (Repository r : repositories) {
            if (!r.isBare()) {
                IPath repoPath = new Path(r.getWorkTree().getAbsolutePath());
                if (location != null && repoPath.isPrefixOf(location)) {
                    if (repository == null || repoPath.segmentCount() > largestSegmentCount) {
                        repository = r;
                        largestSegmentCount = repoPath.segmentCount();
                    }
                }
            }
        }
        return repository;
    }

    public static synchronized void init() throws IOException {
        final List<String> repoPaths = new ArrayList<String>(RepositoryUtil.getRepositoryPaths());
        for (String path : repoPaths) {
            File normalizedGitDir = new Path(new File(path).getAbsolutePath()).toFile();
            Repository r = FileRepositoryBuilder.create(normalizedGitDir);
            repositoryCache.add(r);
        }
    }
}
