package fse.eclipse.mergehelper.detector.merge;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.jtool.changerepository.data.FileInfo;
import org.jtool.changerepository.data.ProjectInfo;
import org.jtool.changerepository.operation.UnifiedOperation;
import org.jtool.changerepository.parser.CodeRange;
import org.jtool.changerepository.parser.OpJavaElement;

import fse.eclipse.mergehelper.detector.AbstractDetector;
import fse.eclipse.mergehelper.detector.DetectOption;
import fse.eclipse.mergehelper.element.BranchInfo;
import fse.eclipse.mergehelper.element.BranchJavaElement;
import fse.eclipse.mergehelper.element.BranchRootInfo;
import fse.eclipse.mergehelper.element.ConflictInfo;
import fse.eclipse.mergehelper.element.MergePoint;
import fse.eclipse.mergehelper.element.MergeType;
import fse.eclipse.mergehelper.util.Parser;

public class AMergePointSearcher extends AbstractDetector {
    private static final String TASK_NAME = "Search Artificial Merge Point";
    private static final String ERROR_MESSAGE = "Artificial Merge Point was Not Found";
    private static AbstractDetector instance = new AMergePointSearcher();

    private boolean isAMerge;
    private MergePoint mPoint;
    private BranchJavaElement a_elemBase, j_elemBase;
    private int a_limitIdx, j_limitIdx;

    private AMergePointSearcher() {
    }

    public static AbstractDetector getInstance() {
        return instance;
    }

    @Override
    public void execute() {
        isAMerge = DetectOption.getInstance().isAMerge();
        if (isAMerge) {
            BranchRootInfo rootInfo = BranchRootInfo.getInstance();

            BranchInfo a_bInfo = rootInfo.getBranchInfo(MergeType.ACCEPT);
            BranchInfo j_bInfo = rootInfo.getBranchInfo(MergeType.JOIN);
            ConflictInfo cInfo = rootInfo.getConflictInfo();

            ProjectInfo a_pInfo = a_bInfo.getProjectInfo();
            ProjectInfo j_pInfo = j_bInfo.getProjectInfo();

            mPoint = new MergePoint();
            searchBeforeConflictElementPoint(cInfo, a_pInfo, j_pInfo);
            if (mPoint.size() == 0 || !a_elemBase.equalsFileElement(j_elemBase)) {
                rootInfo.setMergePoint(mPoint);
                return;
            }

            searchEqualBodyLimitIndex(cInfo, a_pInfo, j_pInfo);
            searchEqualBodyPoint(cInfo, a_pInfo, j_pInfo);

            mPoint.sort();
            rootInfo.setMergePoint(mPoint);
        }
    }

    private void searchBeforeConflictElementPoint(ConflictInfo cInfo, ProjectInfo a_pInfo, ProjectInfo j_pInfo) {
        UnifiedOperation a_operation = null, j_operation = null;

        Map<BranchJavaElement, BranchJavaElement> elemMap = cInfo.getConflictElementMap();
        for (Entry<BranchJavaElement, BranchJavaElement> entry : elemMap.entrySet()) {
            BranchJavaElement a_elem = entry.getKey();
            BranchJavaElement j_elem = entry.getValue();
            UnifiedOperation a_op = a_elem.getOperations().get(0);
            UnifiedOperation j_op = j_elem.getOperations().get(0);

            int idx = a_op.indexOfProjectInfo();
            if (a_operation == null || a_operation.indexOfProjectInfo() - idx > 0) {
                a_operation = a_pInfo.getOperation(idx - 1);
                a_elemBase = a_elem;
            }

            idx = j_op.indexOfProjectInfo();
            if (j_operation == null || j_operation.indexOfProjectInfo() - idx > 0) {
                j_operation = j_pInfo.getOperation(idx - 1);
                j_elemBase = j_elem;
            }
        }

        if (a_operation != null && j_operation != null) {
            mPoint.addPoint(a_operation, j_operation);
        }
    }

    private void searchEqualBodyLimitIndex(ConflictInfo cInfo, ProjectInfo a_pInfo, ProjectInfo j_pInfo) {
        Map<BranchJavaElement, BranchJavaElement> elemMap = cInfo.getConflictElementMap();
        List<UnifiedOperation> a_ops = a_pInfo.getOperations();
        List<UnifiedOperation> j_ops = j_pInfo.getOperations();
        int a_size = a_ops.size();
        int j_size = j_ops.size();

        int a = mPoint.getOperation(0, MergeType.ACCEPT).indexOfProjectInfo() + 1;
        int j = mPoint.getOperation(0, MergeType.JOIN).indexOfProjectInfo() + 1;
        for (; a < a_size && j < j_size; a++, j++) {
            UnifiedOperation a_op = a_ops.get(a);
            UnifiedOperation j_op = j_ops.get(j);

            for (Entry<BranchJavaElement, BranchJavaElement> entry : elemMap.entrySet()) {
                BranchJavaElement a_elem = entry.getKey();
                if (!a_elem.equalsFileElement(a_elemBase) && a_elem.contains(a_op)) {
                    a_limitIdx = a;
                    j_limitIdx = j;
                    return;
                }

                BranchJavaElement j_elem = entry.getValue();
                if (!j_elem.equalsFileElement(j_elemBase) && j_elem.contains(j_op)) {
                    a_limitIdx = a;
                    j_limitIdx = j;
                    return;
                }
            }
        }

        a_limitIdx = a_size;
        j_limitIdx = j_size;
    }

    private void searchEqualBodyPoint(ConflictInfo cInfo, ProjectInfo a_pInfo, ProjectInfo j_pInfo) {
        List<UnifiedOperation> a_ops = a_pInfo.getOperations();
        FileInfo a_fInfo = a_elemBase.getBranchFileInfo().getFileInfo();
        Map<UnifiedOperation, String> a_bodyMap = new LinkedHashMap<>();
        int a = mPoint.getOperation(0, MergeType.ACCEPT).indexOfProjectInfo() + 1;
        for (; a < a_limitIdx; a++) {
            UnifiedOperation a_op = a_ops.get(a);
            if (a_elemBase.contains(a_op)) {
                String body = collectElementBody(a_fInfo, a_op);
                a_bodyMap.put(a_op, body);
            }
        }

        List<UnifiedOperation> j_ops = j_pInfo.getOperations();
        FileInfo j_fInfo = j_elemBase.getBranchFileInfo().getFileInfo();
        Map<UnifiedOperation, String> j_bodyMap = new LinkedHashMap<>();
        int j = mPoint.getOperation(0, MergeType.JOIN).indexOfProjectInfo() + 1;
        for (; j < j_limitIdx; j++) {
            UnifiedOperation j_op = j_ops.get(j);
            if (j_elemBase.contains(j_op)) {
                String body = collectElementBody(j_fInfo, j_op);
                if (body != null) {
                    j_bodyMap.put(j_op, body);
                }
            }
        }

        for (UnifiedOperation a_op : a_bodyMap.keySet()) {
            String a_body = a_bodyMap.get(a_op);
            for (UnifiedOperation j_op : j_bodyMap.keySet()) {
                String j_body = j_bodyMap.get(j_op);
                if (a_body.equals(j_body)) {
                    mPoint.addPoint(a_op, j_op);
                }
            }
        }
    }

    private String collectElementBody(FileInfo fInfo, UnifiedOperation op) {
        String code = fInfo.getCode(op.indexOfFileInfo());
        List<OpJavaElement> elems = new ArrayList<>();
        boolean parseable = Parser.collectElements(fInfo, code, elems, true);
        if (!parseable) {
            throw new Error();
        }

        int start = op.getStart();
        for (OpJavaElement elem : elems) {
            CodeRange range = elem.getCodeRange();
            if (range.inRange(start)) {
                return code.substring(range.getStart(), range.getEnd());
            }
        }
        return null;
    }

    @Override
    protected String getTaskName() {
        return TASK_NAME;
    }

    @Override
    protected String getErrorMessage() {
        return ERROR_MESSAGE;
    }

    @Override
    protected void nextState() {
        if (isAMerge) {
            if (mPoint.size() > 0) {
                AMerge.getInstance().detect(dialog);
            } else {
                error();
            }
        } else {
            finish();
        }
    }
}
